@@ -2,14 +2,15 @@
 
 # Build zipfiles for Pango on Win32: separate runtime and developer packages
 
-ZIP=/tmp/pango-1.1.1-`date +%Y%m%d`.zip
-DEVZIP=/tmp/pango-dev-1.1.1-`date +%Y%m%d`.zip
+ZIP=/tmp/pango-1.1.2-`date +%Y%m%d`.zip
+DEVZIP=/tmp/pango-dev-1.1.2-`date +%Y%m%d`.zip
 cd /usr
 
 rm $ZIP
-zip -r $ZIP -@ <<EOF
+zip $ZIP -@ <<EOF
 bin/pango-querymodules.exe
-etc/pango
+etc/pango/pango.aliases
+etc/pango/pango.modules
 lib/libpango-1.0-0.dll
 lib/libpangoft2-1.0-0.dll
 lib/libpangowin32-1.0-0.dll
